from attr import dataclass

from ros._literals import IPProtocol, MACProtocol


@dataclass(slots=True)
class Torch:
    section: int
    rx: int
    rx_packets: int
    tx: int
    tx_packets: int
    cpu: int = None
    dscp: int = None
    src_port: str = None
    dst_port: str = None
    ip_protocol: IPProtocol = None
    mac_protocol: MACProtocol = None
    src_address: str = None
    dst_address: str = None



from attr import dataclass


@dataclass(slots=True)
class Logging:
    id: str
    action: str
    default: bool
    disabled: bool
    invalid: bool
    prefix: str
    topics: str

    def __bool__(self) -> bool:
        return not self.disabled



from attr import dataclass


@dataclass
class LDPRemoteMapping:
    label: str
    nexthop: str = None
    dst_address: str = None
    vrf: str = None
    comment: str = None
    disabled: bool = None
    copy_from: str = None
    id: str = None
    dynamic: bool = None
    inactive: bool = None
    path: str = None
    peer: str = None



from attr import dataclass
from typing import Literal
from ros._literals import ARPLiteral


@dataclass
class Vrrp:
    name: str
    interface: str
    mtu: int = 1500
    arp: ARPLiteral = "enabled"
    arp_timeout: str = None
    vrid: int = 1
    priority: int = 100
    group_authority: str = "none"
    interval: str = "1.00"
    preemption_mode: bool = True
    authentication: Literal["none", "simple", "ah"] = "none"
    version: Literal["2", "3"] = "2"
    v3_protocol: Literal["ipv4", "ipv6"] = "ipv4"
    sync_connection_tracking: bool = False
    remote_address: str = None
    password: str = None
    on_master: str = None
    on_backup: str = None
    on_fail: str = None
    comment: str = None
    disabled: bool = None
    copy_from: str = None
    id: str = None
    grp_member: bool = None
    mac_address: str = None
    running: bool = None
    invalid: bool = None

    def __str__(self) -> str:
        return self.name



from attr import dataclass


@dataclass
class Wireguard:
    name: str
    mtu: int = 1420
    listen_port: int = 13231
    private_key: str = None
    comment: str = None
    disabled: str = False
    copy_from: str = None

    def __str__(self) -> str:
        return self.name



from attr import dataclass, field
from typing import Literal

AddDefaultRoute = Literal["yes", "no", "special-classless"]
Status = Literal[
    "bound",
    "error",
    "rebinding...",
    "requesting...",
    "searching...",
    "stopped",
    "renewing...",
]


@dataclass
class DHCPClient:
    interface: str
    add_default_route: AddDefaultRoute = "yes"
    dhcp_options: str = "hostname,clientid"
    disabled: bool = False
    use_peer_dns: bool = True
    use_peer_ntp: bool = True
    default_route_distance: int = 1
    comment: str = None
    address: str = None
    dhcp_server: str = field(on_setattr=None, default=None)
    expires_after: str = field(on_setattr=None, default=None)
    gateway: str = field(on_setattr=None, default=None)
    primary_dns: str = field(on_setattr=None, default=None)
    primary_ntp: str = field(on_setattr=None, default=None)
    secondary_dns: str = field(on_setattr=None, default=None)
    status: Status = field(on_setattr=None, default=None)
    dynamic: bool = field(on_setattr=None, default=None)
    invalid: bool = field(on_setattr=None, default=None)
    id: str = field(on_setattr=None, default=None)

    def __str__(self) -> str:
        return self.interface

    def __bool__(self) -> bool:
        return not self.disabled



from attr import dataclass


@dataclass
class BFDSession:
    multihop: bool
    vrf: str
    remote_address: bool
    local_address: bool
    state: str
    state_changes: int
    desired_tx_interval: str
    actual_tx_interval: str
    required_min_rx: str
    remote_min_rx: str
    multiplier: int
    packets_rx: int
    packets_tx: int
    uptime: str = None
    remote_min_tx: str = None
    hold_time: str = None
    up: bool = None



from attr import dataclass


@dataclass
class BridgePort:
    nextid: str
    auto_isolate: bool
    bpdu_guard: bool
    bridge: str
    broadcast_flood: bool
    disabled: bool
    dynamic: bool
    edge: str
    fast_leave: bool
    frame_types: str
    horizon: str
    inactive: bool
    ingress_filtering: bool
    interface: str
    internal_path_cost: int
    learn: str
    multicast_router: str
    path_cost: int
    point_to_point: str
    priority: str
    pvid: int
    restricted_role: bool
    restricted_tcn: bool
    status: str
    tag_stacking: bool
    trusted: bool
    unknown_multicast_flood: bool
    unknown_unicast_flood: bool
    comment: str = None
    id: str = None

    def __bool__(self) -> bool:
        return not self.disabled



from attr import dataclass
from typing import Literal

ServiceName = Literal[
    "api", "api-ssl", "ftp", "ssh", "telnet", "winbox", "www", "www-ssl"
]


@dataclass
class Service:
    id: str
    name: ServiceName
    address: str
    port: int
    disabled: bool
    invalid: bool
    vrf: str = None
    certificate: str = None
    tls_version: str = None

    def __str__(self) -> str:
        return self.name



from attr import dataclass
from typing import Union


@dataclass
class HotspotServer:
    name: str
    interface: str
    address_pool: str = "none"
    profile: str = "default"
    idle_timeout: str = "00:05:00"
    keepalive_timeout: str = None
    login_timeout: str = None
    addresses_per_mac: Union[int, str] = None
    copy_from: str = None
    disabled: bool = None
    https: bool = False
    proxy_status: str = None
    id: str = None
    invalid: bool = None

    def __str__(self) -> str:
        return self.name



from attr import dataclass


@dataclass
class LDPInterface:
    interface: str
    hello_interval: str = None
    hold_time: str = None
    transport_addresses: str = None
    accept_dynamic_neighbors: bool = None
    afi: str = None
    comment: str = None
    disabled: bool = None
    copy_from: str = None
    id: str = None



from attr import dataclass


@dataclass
class User:
    name: str
    group: str
    address: str
    disabled: bool
    expired: bool = False
    id: str = None
    last_logged_in: str = None

    def __str__(self) -> str:
        return self.name



from ros._base import BaseModule, BaseProps

from .address import Address
from .arp import ARP
from .cloud import Cloud
from .dhcp_client import DHCPClient
from .dhcp_relay import DHCPRelay
from .dhcp_server import DHCPServerModule, DHCPServer
from .dns import DNS
from .firewall import IPFirewallModule
from .hotspot import HotspotModule, HotspotServer
from .kid_control import KidControlModule, KidControl
from .neighbor import IPNeighbor
from .pool import IPPool
from .route import Route
from .service import Service
from .setting import Setting
from .vrf import Vrf


class IPModule(BaseModule):
    _dhcp_server: DHCPServerModule = None
    _firewall: IPFirewallModule = None
    _hotspot: HotspotModule = None

    _address: BaseProps[Address] = None
    _arp: BaseProps[ARP] = None
    _dhcp_client: BaseProps[DHCPClient] = None
    _dhcp_relay: BaseProps[DHCPRelay] = None
    _kid_control: KidControlModule = None
    _pool: BaseProps[IPPool] = None
    _neighbor: BaseProps[IPNeighbor] = None
    _route: BaseProps[Route] = None
    _service: BaseProps[Service] = None
    _vrf: BaseProps[Vrf] = None

    @property
    def address(self):
        if not self._address:
            self._address = BaseProps(self.ros, "/ip/address", Address)
        return self._address

    @property
    def arp(self):
        if not self._arp:
            self._arp = BaseProps(self.ros, "/ip/arp", ARP)
        return self._arp

    @property
    def cloud(self) -> Cloud:
        return self.ros.get_as("/ip/cloud", Cloud)

    @property
    def dhcp_client(self):
        if not self._dhcp_client:
            self._dhcp_client = BaseProps(self.ros, "/ip/dhcp-client", DHCPClient)
        return self._dhcp_client

    @property
    def dhcp_relay(self):
        if not self._dhcp_relay:
            self._dhcp_relay = BaseProps(self.ros, "/ip/dhcp-relay", DHCPRelay)
        return self._dhcp_relay

    @property
    def dhcp_server(self) -> DHCPServerModule:
        if not self._dhcp_server:
            self._dhcp_server = DHCPServerModule(
                self.ros, "/ip/dhcp-server", DHCPServer
            )
        return self._dhcp_server

    @property
    def dns(self) -> DNS:
        dns = self.ros.get_as("/ip/dns", DNS)
        dns._ros = self.ros
        return dns

    @property
    def firewall(self) -> IPFirewallModule:
        if not self._firewall:
            self._firewall = IPFirewallModule(self.ros, "/ip/firewall")
        return self._firewall

    @property
    def hotspot(self) -> HotspotModule:
        if not self._hotspot:
            self._hotspot = HotspotModule(self.ros, "/ip/hotspot", HotspotServer)
        return self._hotspot

    @property
    def kid_control(self) -> KidControlModule:
        if not self._kid_control:
            self._kid_control = KidControlModule(
                self.ros, "/ip/kid-control", KidControl
            )
        return self._kid_control

    @property
    def neighbor(self) -> BaseProps[IPNeighbor]:
        if not self._neighbor:
            self._neighbor = BaseProps(self.ros, "/ip/neighbor", IPNeighbor)
            self._neighbor._delete = False
            self._neighbor._write = False
            self._neighbor._disable = False
        return self._neighbor

    @property
    def pool(self) -> BaseProps[IPPool]:
        if not self._pool:
            self._pool = BaseProps(self.ros, "/ip/pool", IPPool)
            self._pool._disable = False
        return self._pool

    @property
    def route(self) -> BaseProps[Route]:
        if not self._route:
            self._route = BaseProps(self.ros, "/ip/route", Route)
        return self._route

    @property
    def service(self) -> BaseProps[Service]:
        if not self._service:
            self._service = BaseProps(self.ros, "/ip/service", Service)
            self._service._create = False
            self._service._delete = False
        return self._service

    @property
    def setting(self) -> Setting:
        return self.ros.get_as("/ip/setting", Setting)

    @property
    def vrf(self) -> BaseProps[Vrf]:
        if not self._vrf:
            self._vrf = BaseProps(self.ros, "/ip/vrf", Vrf)
        return self._vrf


__all__ = [
    "Address",
    "ARP",
    "IPModule",
    "Cloud",
    "DHCPClient",
    "DHCPRelay",
    "DHCPServer",
    "DNS",
    "KidControl",
    "Route",
    "Service",
    "Setting",
    "Vrf",
]



from attr import dataclass


@dataclass
class KidControl:
    name: str
    sun: str = None
    mon: str = None
    tue: str = None
    wed: str = None
    thu: str = None
    fri: str = None
    sat: str = None
    rate_limit: str = None
    tur_sun: str = None
    tur_mon: str = None
    tur_tue: str = None
    tur_wed: str = None
    tur_thu: str = None
    tur_fri: str = None
    tur_sat: str = None
    disabled: bool = None
    blocked: bool = None
    id: str = None
    copy_from: str = None

    def __str__(self) -> str:
        return self.name



from ros._base import BaseProps

from .environment import Environment
from .job import Job
from .script import Script


class ScriptModule(BaseProps[Script]):
    _environment: BaseProps[Environment] = None
    _job: BaseProps[Job] = None

    @property
    def job(self):
        if not self._job:
            self._job = BaseProps(self.ros, "/system/script/job", Job)
            self._job._create = False
            self._job._write = False
        return self._job

    @property
    def environment(self):
        if not self._environment:
            self._environment = BaseProps(
                self.ros, "/system/script/environment", Environment
            )
        return self._environment

    def run(self, script: Script):
        return self.ros.post_as("/system/script/run", None, {".id": script.id})


__all__ = ["Environment", "Job", "Script", "ScriptModule"]



from ros._base import BaseModule, BaseProps

from .bfd import BFDModule
from .id import RoutingId
from .nexthop import Nexthop
from .rule import RoutingRule
from .table import RoutingTable


class RoutingModule(BaseModule):
    _bfd: BFDModule = None
    _id: BaseProps[RoutingId] = None
    _nexthop: BaseProps[Nexthop] = None
    _rule: BaseProps[RoutingRule] = None
    _table: BaseProps[RoutingTable] = None

    @property
    def bfd(self):
        if not self._bfd:
            self._bfd = BFDModule(self.ros, "/routing/bfd")
        return self._bfd

    @property
    def id(self):
        if not self._id:
            self._id = BaseProps(self.ros, "/routing/id", RoutingId)
        return self._id

    @property
    def nexthop(self):
        if not self._nexthop:
            self._nexthop = BaseProps(self.ros, "/routing/nexthop", Nexthop)
        return self._nexthop

    @property
    def rule(self):
        if not self._rule:
            self._rule = BaseProps(self.ros, "/routing/rule", RoutingRule)
        return self._rule

    @property
    def table(self):
        if not self._table:
            self._table = BaseProps(self.ros, "/routing/table", RoutingTable)
        return self._table


__all__ = ["RoutingId", "Nexthop", "RoutingModule", "RoutingTable"]



from attr import dataclass
from typing import Literal

QueueKind = Literal[
    "bfifo",
    "cake",
    "codel",
    "fq-codel",
    "mq-pfifo",
    "none",
    "pcq",
    "pfifo",
    "red",
    "sfq",
]
CakeAckFilter = Literal["aggressive", "filter", "none"]
CakeAtm = Literal["atm", "ptm", "none"]
CakeDiffserv = Literal[
    "besteffort", "diffserv3", "diffserv4", "diffserv8", "precedence"
]
CakeFlowmode = Literal[
    "dsthost",
    "dual-dsthost",
    "dual-srchost",
    "flowblind",
    "flows",
    "hosts",
    "srchost",
    "triple-isolate",
]
CakeOverheadScheme = Literal[
    "bridged-llcsnap",
    "bridged-ptm",
    "bridged-vcmux",
    "convervative",
    "docsis",
    "ether-vlan",
    "ethernet",
    "ipoa-llcsnap",
    "ipoa-vcmux",
    "pppoa-llc",
    "pppoa-vcmux",
    "pppoe-llcsnap",
    "pppoe-ptm",
    "pppoe-vcmux",
    "raw",
    "via-ethernet",
    "",
]
CakeRttScheme = Literal[
    "datacentre",
    "internet",
    "interplanetary",
    "lan",
    "metro",
    "none",
    "oceanic",
    "regional",
    "satellite",
]


@dataclass
class QueueType:
    name: str
    kind: QueueKind
    bfifo_limit: int = None
    cake_ack_filter: CakeAckFilter = None
    cake_atm: CakeAtm = None
    cake_autorate_ingress: bool = None
    cake_bandwidth: str = None
    cake_diffserv: CakeDiffserv = None
    cake_flowmode: CakeFlowmode = None
    cake_memlimit: str = None
    cake_mpu: int = None
    cake_nat: bool = None
    cake_overhead: int = None
    cake_overhead_scheme: CakeOverheadScheme = None
    cake_rtt: str = None
    cake_rtt_scheme: CakeRttScheme = None
    cake_wash: bool = None
    codel_ce_threshold: int = None
    codel_ecn: bool = None
    codel_interval: str = None
    codel_limit: int = None
    codel_target: str = None
    fq_codel_ce_threshold: int = None
    fq_codel_ecn: bool = None
    fq_codel_flows: int = None
    fq_codel_interval: str = None
    fq_codel_limit: int = None
    fq_codel_memlimit: int = None
    fq_codel_quantum: int = None
    fq_codel_target: str = None
    mq_pfifo_limit: int = None
    pcq_burst_rate: int = None
    pcq_burst_threshold: int = None
    pcq_burst_time: str = None
    pcq_classifier: str = None
    pcq_dst_address_mask: int = None
    pcq_dst_address6_mask: int = None
    pcq_limit: int = None
    pcq_rate: int = None
    pcq_src_address_mask: int = None
    pcq_src_address6_mask: int = None
    pcq_total_limit: int = None
    pfifo_limit: int = None
    red_avg_packet: int = None
    red_burst: int = None
    red_limit: int = None
    red_max_threshold: int = None
    red_min_threshold: int = None
    sfq_allot: int = None
    sfq_perturb: int = None
    copy_from: str = None



from attr import dataclass
from typing import Literal

from ._literals import Authoritative, UseRadius


@dataclass
class DHCPServer:
    name: str
    interface: str
    address_pool: str = "static-only"
    lease_time: str = "00:10:00"
    authoritative: Authoritative = "yes"
    disabled: bool = False
    use_radius: UseRadius = "no"
    always_broadcast: bool = False
    add_arp: bool = False
    conflict_detection: bool = True
    use_framed_as_classless: bool = True
    use_radius: bool = "no"
    lease_script: str = None
    insert_queue_before: str = "first"
    parent_queue: str = None
    allow_dual_stack_queue: bool = True
    comment: str = None
    relay: str = None
    client_mac_limit: str = None
    bootp_lease_time: str = "forever"
    delay_threshold: str = None
    dhcp_option_set: str = None
    server_address: str = None
    bootp_support: Literal["dynamic", "static"] = None
    invalid: bool = None
    dynamic: bool = None
    id: str = None

    def __str__(self) -> str:
        return self.name

    def __bool__(self) -> bool:
        return not self.disabled



from attr import dataclass

from ._literals import DNSType


@dataclass
class DNSStatic:
    id: str
    disabled: bool
    ttl: str
    dynamic: bool = False
    address: str = None
    cname: str = None
    comment: str = None
    forward_to: str = None
    ipv6_address: str = None
    name: str = None
    ns: str = None
    mx_preference: str = None
    mx_exchange: str = None
    regexp: str = None
    srv_priority: int = None
    srv_weight: int = None
    srv_port: int = None
    srv_target: str = None
    text: str = None
    type: DNSType = None

    def __bool__(self) -> bool:
        return not self.disabled



from attr import dataclass


@dataclass
class LDPAdvertiseFilter:
    vrf: str
    prefix: str
    neighbor: str = None
    advertise: bool = None
    comment: str = None
    copy_from: str = None
    disabled: bool = None
    place_before: str = None
    id: str = None
    nextid: str = None



from attr import dataclass


@dataclass
class PPPAAA:
    accounting: bool
    interim_update: str
    use_circuit_id_in_nas_port_id: bool
    use_radius: bool
    enable_ipv6_accounting: bool = None



from typing import Any, List

from ros._base import BaseModule, BaseProps

from .bridge import BridgeModule, Bridge, BridgeMsti, BridgePort, BridgeVlan
from .eoip import EoIP
from .ethernet import InterfaceEthernet
from .interface import Interface
from .list import InterfaceList, InterfaceListMember, InterfaceListModule
from .veth import Veth
from .vlan import Vlan
from .wireguard import WireguardModule, Wireguard
from .vrrp import Vrrp


class InterfaceModule(BaseModule):
    _brigde: BridgeModule = None
    _eoip: BaseProps[EoIP] = None
    _ethernet: BaseProps[InterfaceEthernet] = None
    _list: InterfaceListModule = None
    _veth: BaseProps[Veth] = None
    _vlan: BaseProps[Vlan] = None
    _vrrp: BaseProps[Vrrp] = None
    _wireguard: WireguardModule = None

    def __call__(self, **kwds: Any):
        return self.print(**kwds)

    @property
    def bridge(self) -> BridgeModule:
        if not self._brigde:
            self._brigde = BridgeModule(self.ros, "/interface/bridge", Bridge)
        return self._brigde

    @property
    def list(self) -> InterfaceListModule:
        if not self._list:
            self._list = InterfaceListModule(self.ros, "/interface/list", InterfaceList)
        return self._list

    @property
    def eoip(self) -> BaseProps[EoIP]:
        if not self._eoip:
            self._eoip = BaseProps(self.ros, "/interface/eoip", EoIP)
        return self._eoip

    @property
    def ethernet(self) -> BaseProps[InterfaceEthernet]:
        if not self._ethernet:
            self._ethernet = BaseProps(
                self.ros, "/interface/ethernet", InterfaceEthernet
            )
            self._ethernet._create = False
            self._ethernet._delete = False
        return self._ethernet

    @property
    def veth(self) -> BaseProps[Veth]:
        if not self._veth:
            self._veth = BaseProps(self.ros, "/interface/veth", Veth)
        return self._veth

    @property
    def vlan(self) -> BaseProps[Vlan]:
        if not self._vlan:
            self._vlan = BaseProps(self.ros, "/interface/vlan", Vlan)
        return self._vlan

    @property
    def vrrp(self) -> BaseProps[Vrrp]:
        if not self._vrrp:
            self._vrrp = BaseProps(self.ros, "/interface/vrrp", Vrrp)
        return self._vrrp

    @property
    def wireguard(self) -> WireguardModule:
        if not self._wireguard:
            self._wireguard = WireguardModule(
                self.ros, "/interface/wireguard", Wireguard
            )
        return self._wireguard

    def print(self, **kwds: Any) -> List[Interface]:
        return self.ros.get_as(self.url, List[Interface], kwds)


__all__ = [
    "BridgeModule",
    "Bridge",
    "BridgeMsti",
    "BridgePort",
    "BridgeVlan",
    "EthernetModule",
    "InterfaceModule",
    "Interface",
    "InterfaceList",
    "InterfaceListMember",
    "InterfaceListModule",
]



from attr import dataclass


@dataclass
class Script:
    name: str
    source: str
    dont_require_permissions: bool
    policy: str
    comment: str = None
    id: str = None
    nextid: str = None
    copy_from: str = None
    owner: str = None
    run_count: int = None
    last_started: str = None
    invalid: bool = None



from attr import dataclass


@dataclass
class BFDConfiguration:
    forbid_bfd: bool
    interfaces: str = None
    addresses: str = None
    address_list: str = None
    min_rx: str = None
    min_tx: str = None
    multiplier: int = None
    comment: str = None
    disabled: bool = False
    vrf: str = None
    id: str = None
    nextid: str = None
    inactive: bool = None
    copy_from: str = None
    place_before: str = None



__version__ = "0.12.2"
from .inteface import InterfaceModule
from .ip import IPModule
from .mpls import MPLSModule
from .ppp import PPPModule
from .queue import QueueModule
from .routing import RoutingModule
from .system import SystemModule
from .tool import ToolModule
from .user import UserModule

from .error import Error
from .log import Log


from .ros import Ros

__all__ = [
    "Error",
    "InterfaceModule",
    "IPModule",
    "MPLSModule",
    "PPPModule",
    "QueueModule",
    "RoutingModule",
    "Log",
    "Ping",
    "Ros",
    "SystemModule",
    "ToolModule",
    "UserModule",
]



from attr import dataclass
from typing import Optional


@dataclass
class UserSSHKey:
    user: str
    bits: str
    key_owner: str
    id: str
    RSA: Optional[bool] = None

    def __str__(self) -> str:
        return self.key_owner or self.user



from attr import dataclass


@dataclass(slots=True)
class Cloud:
    ddns_enabled: bool
    ddns_update_interval: str = None
    update_time: bool = None
    public_address: str = None
    status: str = None
    back_to_home_vpn: str = None

    def __bool__(self) -> bool:
        return not self.ddns_enabled



from attr import dataclass


@dataclass
class MPLSInterface:
    interface: str
    mpls_mtu: int = None
    input: bool = None
    comment: str = None
    copy_from: str = None
    disabled: bool = None
    place_before: str = None
    id: str = None



from attr import dataclass


@dataclass
class MPLSForwardingTable:
    id: str
    label: str
    type: str
    ldp: bool = False
    vpls: bool = False
    vrf: str = None
    nexthops: str = None
    prefix: str = None



from attr import dataclass


@dataclass
class RoutingId:
    id: str
    name: str
    select_from_vrf: str
    select_dynamic_id: str
    disabled: bool = False
    comment: str = None
    copy_from: str = None

    def __str__(self) -> str:
        return self.name



from attr import dataclass

from ros._literals import IPProtocol, TCPState


@dataclass(slots=True)
class IPConnection:
    id: str
    assured: bool
    confirmed: bool
    dst_address: str
    dstnat: bool
    dying: bool
    expected: bool
    fasttrack: bool
    hw_offload: bool
    orig_bytes: int
    orig_fasttrack_bytes: int
    orig_fasttrack_packets: int
    orig_packets: int
    orig_rate: int
    protocol: IPProtocol
    repl_bytes: int
    repl_fasttrack_bytes: int
    repl_fasttrack_packets: int
    repl_packets: int
    repl_rate: int
    reply_dst_address: str
    reply_src_address: str
    seen_reply: bool
    src_address: str
    srcnat: bool
    timeout: str
    connection_mark: str = None
    tcp_state: TCPState = None



from cattrs import Converter
from typing import Any, Dict, Union


def clean_key(d: Dict[str, Any]) -> dict:
    """Replace - with _ and remove . from key

    Args:
        d (Dict[str, Any]): Raw dict data from ros

    Returns:
        dict: Cleaned data ready to be used as python object props value
    """
    nd = dict()
    for k, v in d.items():
        k = k.replace("-", "_")
        k = k.replace(".", "")
        nd[k] = v
    return nd


def clean_data(d: Dict[str, Any]) -> dict:
    """Clean raw data from ros recursively

    Args:
        d (Dict[str, Any]): Data from ros

    Returns:
        dict: Cleaned data
    """
    data: Any = None
    if isinstance(d, dict):
        data = clean_key(d)
    elif isinstance(d, list):
        data = list()
        for val in d:
            if isinstance(val, dict):
                data.append(clean_key(val))
            else:
                data.append(val)
    return data


def clean_filters(d: Dict[str, Any]) -> Dict[str, Any]:
    """Replace _ with - on every key on data for query in url

    Args:
        d (Dict[str, Any]): Data to be clean

    Returns:
        Dict[str, Any]: Cleaned data
    """
    if not d:
        return d
    translation_table = str.maketrans("_", "-")
    return {k.translate(translation_table): v for k, v in d.items()}


def clean_before_put(d: Dict[str, Any]) -> dict:
    """Replace _ with - on every key on data for PUT request

    Args:
        d (Dict[str, Any]): Data to be clean

    Returns:
        dict: cleaned data
    """
    if not d:
        return d
    nd = {k.replace("_", "-"): v for k, v in d.items() if v is not None}
    nd.pop("id", None)
    return nd


def _union_str_int(v: str, t: Any) -> Union[str, int]:
    if v.isdigit():
        return int(v)
    return v


def make_converter() -> Converter:
    """Making converter for cattrs

    Returns:
        Converter: cattrs converter
    """
    c = Converter()
    # Main
    c.register_structure_hook(Union[int, str], _union_str_int)
    return c



from attr import dataclass


@dataclass
class LDPInstance:
    lsr_id: str
    path_vector_limit: str = None
    hop_limit: str = None
    loop_detect: str = None
    use_explicit_null: str = None
    distribute_for_default: str = None
    transport_addresses: str = None
    vrf: str = None
    afi: str = None
    preferred_afi: str = None
    comment: str = None
    disabled: bool = None
    inactive: bool = None
    copy_from: str = None
    id: str = None



from attr import dataclass, field


@dataclass
class ARP:
    address: str
    interface: str
    mac_address: str = None
    published: bool = False
    disabled: bool = False
    comment: str = None
    id: str = field(on_setattr=None, default=None)
    DHCP: bool = field(on_setattr=None, default=None)
    complete: bool = field(on_setattr=None, default=None)
    dynamic: bool = field(on_setattr=None, default=None)
    invalid: bool = field(on_setattr=None, default=None)

    def __bool__(self) -> bool:
        return not self.disabled



from attr import dataclass


@dataclass(slots=True)
class RoutingTable:
    id: str
    name: str
    invalid: bool
    dynamic: bool = False
    fib: str = "no"
    comment: str = None

    def __str__(self) -> str:
        return self.name

    def __attrs_post_init__(self):
        if self.fib == "":
            self.fib = "yes"



from attr import dataclass


@dataclass
class Address:
    address: str
    interface: str
    network: str = None
    comment: str = None
    actual_interface: str = None
    disabled: bool = False
    invalid: str = None
    dynamic: bool = None
    id: str = None

    def __str__(self) -> str:
        return self.address

    def __bool__(self) -> bool:
        return not self.disabled



from ros._base import BaseProps

from .accept_filter import LDPAcceptFilter
from .advertise_filter import LDPAdvertiseFilter
from .instance import LDPInstance
from .interface import LDPInterface
from .local_mapping import LDPLocalMapping
from .neighbor import LDPNeighbor
from .remote_mapping import LDPRemoteMapping


class MPLSLDP(BaseProps[LDPInstance]):
    _accept_filter: BaseProps[LDPAcceptFilter] = None
    _advertise_filter: BaseProps[LDPAdvertiseFilter] = None
    _interface: BaseProps[LDPInterface] = None
    _neighbor: BaseProps[LDPNeighbor] = None
    _local_mapping: BaseProps[LDPLocalMapping] = None
    _remote_mapping: BaseProps[LDPLocalMapping] = None

    @property
    def accept_filter(self):
        if not self._accept_filter:
            self._accept_filter = BaseProps(
                self.ros, "/mpls/ldp/accept-filter", LDPAcceptFilter
            )
        return self._accept_filter

    @property
    def advertise_filter(self):
        if not self._advertise_filter:
            self._advertise_filter = BaseProps(
                self.ros, "/mpls/ldp/advertise-filter", LDPAdvertiseFilter
            )
        return self._advertise_filter

    @property
    def interface(self):
        if not self._interface:
            self._interface = BaseProps(self.ros, "/mpls/ldp/interface", LDPInterface)
        return self._interface

    @property
    def local_mapping(self):
        if not self._local_mapping:
            self._local_mapping = BaseProps(
                self.ros, "/mpls/ldp/local-mapping", LDPLocalMapping
            )
        return self._local_mapping

    @property
    def neighbor(self):
        if not self._neighbor:
            self._neighbor = BaseProps(self.ros, "/mpls/ldp/neighbor", LDPNeighbor)
        return self._neighbor

    @property
    def remote_mapping(self):
        if not self._remote_mapping:
            self._remote_mapping = BaseProps(
                self.ros, "/mpls/ldp/remote-mapping", LDPRemoteMapping
            )
        return self._remote_mapping



from attr import dataclass


@dataclass
class Veth:
    name: str
    address: str
    gateway: str
    comment: str = None
    disabled: bool = False
    copy_from: str = None

    def __str__(self) -> str:
        return self.name



from attr import define

from ros._base import BaseModule, BaseProps

from .address_list import IPAddressList
from .connection import IPConnection
from .filter import IPFirewallFilter
from .mangle import IPFirewallMangle
from .nat import IPFirewallNAT


@define
class IPFirewallModule:
    ros: BaseModule
    filename: str
    _address_list: BaseProps[IPAddressList] = None
    _connection: BaseProps[IPConnection] = None
    _filter: BaseProps[IPFirewallFilter] = None
    _mangle: BaseProps[IPFirewallMangle] = None
    _nat: BaseProps[IPFirewallNAT] = None

    @property
    def address_list(self) -> BaseProps[IPAddressList]:
        if not self._address_list:
            self._address_list = BaseProps(
                self.ros, "/ip/firewall/address-list", IPAddressList
            )
        return self._address_list

    @property
    def connection(self) -> BaseProps[IPConnection]:
        if not self._connection:
            self._connection = BaseProps(
                self.ros, "/ip/firewall/connection", IPConnection
            )
        return self._connection

    @property
    def filter(self) -> BaseProps[IPFirewallFilter]:
        if not self._filter:
            self._filter = BaseProps(self.ros, "/ip/firewall/filter", IPFirewallFilter)
        return self._filter

    @property
    def mangle(self) -> BaseProps[IPFirewallMangle]:
        if not self._mangle:
            self._mangle = BaseProps(self.ros, "/ip/firewall/mangle", IPFirewallMangle)
        return self._mangle

    @property
    def nat(self) -> BaseProps[IPFirewallNAT]:
        if not self._nat:
            self._nat = BaseProps(self.ros, "/ip/firewall/nat", IPFirewallNAT)
        return self._nat


__all__ = [
    "IPConnection",
    "IPFirewallFilter",
    "IPFirewallMangle",
    "IPFirewallNAT",
    "IPFirewallModule",
]



from ros._base import BaseModule, BaseProps

from .interface import InterfaceQueue
from .simple import SimpleQueue, SimpleQueueModule
from .tree import QueueTree, QueueTreeModule
from .type import QueueType


class QueueModule(BaseModule):
    _interface: BaseProps[InterfaceQueue] = None
    _simple: SimpleQueueModule = None
    _tree: QueueTreeModule = None
    _type: BaseProps[QueueType] = None

    @property
    def interface(self):
        if not self._interface:
            self._interface = BaseProps(self.ros, "/queue/interface", InterfaceQueue)
        return self._interface

    @property
    def simple(self):
        if not self._simple:
            self._simple = SimpleQueueModule(self.ros, "/queue/simple", SimpleQueue)
        return self._simple

    @property
    def tree(self):
        if not self._tree:
            self._tree = QueueTreeModule(self.ros, "/queue/tree", QueueTree)
        return self._tree

    @property
    def type(self):
        if not self._type:
            self._type = BaseProps(self.ros, "/queue/type", QueueType)
        return self._type



from ros._base import BaseModule, BaseProps

from .health import Health
from .history import History
from .identity import Identity
from .license import License
from .logging import Logging
from .note import Note
from .package import Package, PackageModule
from .resource import Resource
from .routerboard import RouterBOARD
from .script import Script, ScriptModule


class SystemModule(BaseModule):
    _health: BaseProps[Health] = None
    _history: BaseProps[History] = None
    _logging: BaseProps[Logging] = None
    _package: PackageModule = None
    _script: ScriptModule = None

    @property
    def health(self):
        if not self._health:
            self._health = BaseProps(self.ros, "/system/health", Health)
            self._health._write = False
        return self._health

    @property
    def history(self):
        if not self._history:
            self._history = BaseProps(self.ros, "/system/history", History)
            self._history._write = False
            self._history._delete = False
        return self._history

    @property
    def identity(self) -> Identity:
        return self.ros.get_as("/system/identity", Identity)

    @property
    def license(self) -> License:
        return self.ros.get_as("/system/license", License)

    @property
    def logging(self):
        if not self._logging:
            self._logging = BaseProps(self.ros, "/system/logging", Logging)
            self._logging._create = False
        return self._logging

    @property
    def note(self) -> Note:
        return self.ros.get_as("/system/note", Note)

    @property
    def package(self) -> PackageModule:
        if not self._package:
            self._package = PackageModule(self.ros, "/system/package", Package)
            self._package._create = False
            self._package._write = False
        return self._package

    @property
    def resource(self) -> Resource:
        return self.ros.get_as("/system/resource", Resource)

    @property
    def routerboard(self) -> RouterBOARD:
        return self.ros.get_as("/system/routerboard", RouterBOARD)

    @property
    def script(self) -> ScriptModule:
        if not self._script:
            self._script = ScriptModule(self.ros, "/system/script", Script)
        return self._script

    def ssh_exec(
        self,
        address: str,
        command: str,
        output_to_file: str = None,
        port: int = None,
        src_address: str = None,
        user: str = None,
        vrf: str = None,
    ) -> str:
        data = {"address": address, "command": command}
        return self.ros.post_as("/system/ssh-exec", str, data)


__all__ = [
    "Health",
    "History",
    "Identity",
    "License",
    "Logging",
    "Note",
    "Resource",
    "Script",
    "RouterBOARD",
    "SystemModule",
]



from attr import dataclass
from typing import Literal

ActionLiteral = Literal["accept", "drop", "reject"]


@dataclass
class HotspotWalledGardenIP:
    action: ActionLiteral = "accept"
    server: str = None
    src_address: str = None
    dst_address: str = None
    src_address_list: str = None
    dst_address_list: str = None
    protocol: str = None
    dst_port: str = None
    dst_host: str = None
    comment: str = None
    disabled: bool = False
    id: str = None
    copy_from: str = None
    invalid: bool = False
    place_before: str = None



from attr import dataclass, field
from typing import Union
from ros._base import BaseProps


@dataclass
class SimpleQueue:
    # General
    name: str
    target: str
    max_limit: str = "0/0"
    dst: str = None
    burst_limit: str = None
    burst_threshold: str = None
    burst_time: str = None
    time: str = None
    disabled: bool = False
    comment: str = None
    # Advanced
    packet_marks: str = None
    limit_at: str = "0/0"
    priority: str = "8/8"
    bucket_size: str = "0.100/0.100"
    queue: str = None
    parent: str = None
    # Total
    total_bucket_size: str = None
    total_burst_threshold: str = None
    total_limit_at: str = None
    total_priority: int = None
    total_burst_limit: str = None
    total_burst_time: str = None
    total_max_limit: str = None
    total_queue: str = None
    # Statistics
    rate: str = field(on_setattr=None, default=None)
    packet_rate: str = field(on_setattr=None, default=None)
    queued_bytes: str = field(on_setattr=None, default=None)
    queued_packets: str = field(on_setattr=None, default=None)
    bytes: str = field(on_setattr=None, default=None)
    packets: str = field(on_setattr=None, default=None)
    dropped: str = field(on_setattr=None, default=None)
    pcq_queues: str = field(on_setattr=None, default=None)
    # Total Statistics
    total_rate: str = field(on_setattr=None, default=None)
    total_packet_rate: int = field(on_setattr=None, default=None)
    total_queued_bytes: str = field(on_setattr=None, default=None)
    total_queued_packets: int = field(on_setattr=None, default=None)
    total_bytes: str = field(on_setattr=None, default=None)
    total_packets: int = field(on_setattr=None, default=None)
    total_dropped: int = field(on_setattr=None, default=None)
    total_pcq_queues: str = field(on_setattr=None, default=None)
    # ETC
    id: str = field(on_setattr=None, default=None)
    dynamic: bool = field(on_setattr=None, default=None)
    invalid: bool = field(on_setattr=None, default=None)

    def __str__(self) -> str:
        return self.name

    def __bool__(self) -> bool:
        return not self.disabled


class SimpleQueueModule(BaseProps[SimpleQueue]):
    def reset(self, queue: Union[SimpleQueue, str]):
        rid = queue.id if isinstance(queue, SimpleQueue) else queue
        self.ros.post_as("/queue/simple/reset", None, {".id": rid})

    def reset_counters(self, queue: Union[SimpleQueue, str]):
        rid = queue.id if isinstance(queue, SimpleQueue) else queue
        self.ros.post_as("/queue/simple/reset-counters", None, {".id": rid})

    def reset_counters_all(self):
        self.ros.post_as("/queue/simple/reset-counters-all", None)

    def move(self, queue: Union[SimpleQueue, str], destination: str):
        rid = queue.id if isinstance(queue, SimpleQueue) else queue
        self.ros.post_as(
            "/queue/simple/move", None, {".id": rid, "destination": destination}
        )



from attr import dataclass
from typing import List

from ros._base import BaseProp
from . import UpdateChannel


@dataclass
class Update:
    channel: UpdateChannel
    installed_version: str
    latest_version: str = None
    status: str = None
    section: int = None


class UpdateModule(BaseProp[Update]):
    def cancel(self):
        self.ros.post_as(self.filename + "/cancel", None)

    def check_for_updates(self) -> Update:
        return self.ros.post_as(self.filename + "/check-for-updates", List[Update])

    def download(self):
        self.ros.post_as(self.filename + "/download", None)

    def install(self):
        self.ros.post_as(self.filename + "/install", None)

    def set(self, channel: UpdateChannel):
        return super().set(channel=channel)



from attr import dataclass


@dataclass
class HotspotCookie:
    id: str
    user: str
    mac_address: str
    expires_in: str = None
    domain: str = None

    def __str__(self) -> str:
        return self.user
